#include "lib.h"
#include <inc/fs.h>
#include <stdio.h>

#define debug 0

static int file_close(struct Fd *fd);
static int file_read(struct Fd *fd, void *buf, u_int n, u_int offset);
static int file_write(struct Fd *fd, const void *buf, u_int n, u_int offset);
static int file_stat(struct Fd *fd, struct Stat *stat);

static int fix_dir(const char *, char *);

struct Dev devfile =
{
.dev_id=	'f',
.dev_name=	"file",
.dev_read=	file_read,
.dev_write=	file_write,
.dev_close=	file_close,
.dev_stat=	file_stat,
};


// Open a file (or directory),
// returning the file descriptor on success, < 0 on failure.
static int
_internal_open(const char *path, int mode, int type)
{
  struct Filefd *filefd;
  struct Fd *fd;
  int r, i;
  char real_path[MAXPATHLEN];

  if ((r = fd_alloc(&fd)) < 0) {
    return r;
  }
  if ((r = sys_mem_alloc(0, (u_int)fd, PTE_P|PTE_U|PTE_W)) < 0) {
    return r;
  }

  fix_dir(path, real_path);
  if ((r = fsipc_open(real_path, mode, FTYPE_REG, fd)) < 0){
    if (sys_mem_unmap(0, (u_int)fd) < 0)
      printf("open fsipc_open1 failed to unmap 0x%x\n", fd);
    return r;
  }
  filefd=(struct Filefd *)fd;

  for (i = 0; i < ROUND(filefd->f_file.f_size, BY2PG); i += BY2PG) {
    if ((r = fsipc_map(filefd->f_fileid, i, fd2data((struct Fd *)filefd)+i)) < 0) {
      fsipc_close(filefd->f_fileid);
      if (sys_mem_unmap(0, (u_int)filefd) <0)
	printf("open fsipc_open2 failed to unmap 0x%x\n", filefd);
      return -E_NO_MEM;
    }
  }

  if (mode & O_APPEND) {
    fd->fd_offset = filefd->f_file.f_size;
  }

  return fd2num(fd);
}

// Close a file descriptor
int
file_close(struct Fd *fd)
{
  int i;
  struct Filefd *filefd=(struct Filefd *)fd;

  assert(filefd != NULL);
  for (i = 0; i < ROUND(filefd->f_file.f_size, BY2PG); i += BY2PG) {
    if (sys_mem_unmap(0, fd2data(fd)+i) <0)
      printf("close failed to unmap 0x%x\n", fd2data(fd) + i);
  }
  fsipc_close(filefd->f_fileid);
  if (sys_mem_unmap(0, (u_int)filefd) <0)
    printf("close failed to unmap fd 0x%x\n", filefd);
  return 0;
}

// Read 'n' bytes from 'fd' at the current seek position into 'buf'.
// Since files are memory-mapped, this amounts to a bcopy()
// surrounded by a little red tape to handle the file size and seek pointer.
static int
file_read(struct Fd *fd, void *buf, u_int n, u_int offset)
{
	u_int size;
	struct Filefd *f;

	f = (struct Filefd*)fd;

	// avoid reading past the end of file
	size = f->f_file.f_size;
	if (offset > size)
		return 0;
	if (offset+n > size)
		n = size - offset;

	// read the data by copying from the file mapping
	bcopy((char*)fd2data(fd)+offset, buf, n);
	return n;
}

// Find the virtual address of the page
// that maps the file block starting at 'offset'.
int
read_map(int fdnum, u_int offset, void **blk)
{
	int r;
	u_int va;
	struct Fd *fd;

	if ((r = fd_lookup(fdnum, &fd)) < 0)
		return r;
	if (fd->fd_dev_id != devfile.dev_id)
		return -E_INVAL;
	va = fd2data(fd) + offset;
	if (offset >= MAXFILESIZE)
		return -E_NO_DISK;
	if (!(vpd[PDX(va)]&PTE_P) || !(vpt[VPN(va)]&PTE_P))
		return -E_NO_DISK;
	*blk = (void*)va;
	return 0;
}

// Write 'n' bytes from 'buf' to 'fd' at the current seek position.
static int
file_write(struct Fd *fd, const void *buf, u_int n, u_int offset)
{
	int r;
	u_int tot;
	struct Filefd *f;

	f = (struct Filefd*)fd;

	if (debug) fprintf(STDERR_FILE, "write %d bytes (%d - %d)\n", n, f->f_file.f_size, offset);

	// don't write past the maximum file size
	tot = offset + n;
	if (tot > MAXFILESIZE)
		return -E_NO_DISK;

	// increase the file's size if necessary
	if (tot > f->f_file.f_size) {
		if ((r = ftruncate(fd2num(fd), tot)) < 0)
			return r;
	}

	// write the data
	bcopy(buf, (char*)(fd2data(fd)+offset), n);
	return n;
}

static int
file_stat(struct Fd *fd, struct Stat *st)
{
	struct Filefd *f;

	f = (struct Filefd*)fd;

	strcpy(st->st_name, f->f_file.f_name);
	st->st_size = f->f_file.f_size;
	st->st_isdir = f->f_file.f_type==FTYPE_DIR;
	return 0;
}

// Truncate or extend an open file to 'size' bytes
int
ftruncate(int fdnum, u_int size)
{
	int i, r;
	struct Fd *fd;
	struct Filefd *f;
	u_int oldsize, va, fileid;

	if (size > MAXFILESIZE)
		return -E_NO_DISK;

	if ((r = fd_lookup(fdnum, &fd)) < 0)
		return r;
	if (fd->fd_dev_id != devfile.dev_id)
		return -E_INVAL;

	f = (struct Filefd*)fd;
	fileid = f->f_fileid;
	oldsize = f->f_file.f_size;
	if ((r = fsipc_set_size(fileid, size)) < 0)
		return r;

	va = fd2data(fd);
	// Map any new pages needed if extending the file
	for (i = ROUND(oldsize, BY2PG); i < ROUND(size, BY2PG); i += BY2PG) {
		if ((r = fsipc_map(fileid, i, va+i)) < 0) {
			fsipc_set_size(fileid, oldsize);
			return r;
		}
	}

	// Unmap pages if truncating the file
	for (i = ROUND(size, BY2PG); i < ROUND(oldsize, BY2PG); i+=BY2PG)
		if ((r = sys_mem_unmap(0, va+i)) < 0)
			panic("ftruncate: sys_mem_unmap %08x: %e", va+i, r);
	return 0;
}

// Delete a file
int
remove(const char *path)
{
  char real_path[MAXPATHLEN];

  fix_dir(path, real_path);
  return fsipc_remove(path);
}

// Synchronize disk with buffer cache
int
sync(void)
{
	return fsipc_sync();
}

static char _curdir[MAXNAMELEN+1];

static inline void fix_curdir() {
  char *ptr;
  
  if (*_curdir == '\0') {
    if ((ptr = getenv("CWD")) == NULL) {
      strcpy(_curdir, "/");
    } else {
      strcpy(_curdir, ptr);
    }
  }
}


static int 
fix_dir(const char *src, char *dst)
{
  int slen, cdlen;
  char *ptr;

  fix_curdir();

  if (src == NULL) {
    strcpy(dst, _curdir);
    return 0;
  }

  slen = strlen(src);

  if (*src == '/') {
    if (slen >= MAXNAMELEN) {
      return -E_BAD_PATH;
    }
    strcpy(dst, src);
    return 0;
  } else {
    if (!strcmp(src, "..")) {
      strcpy(dst, _curdir);
      if (dst[strlen(dst)-1] == '/')
	dst[strlen(dst)-1] = '\0';
      if ((ptr = (char *) strrchr(dst, '/')) == 0)
	return -E_BAD_PATH;
      if (dst[1] == '\0')
	return -E_BAD_PATH;
      else {
	if (ptr == dst)
	  strcpy(dst, "/");
	else
	  *ptr = '\0';
	return 0;
      }
    }
    cdlen = strlen(_curdir);
    if (slen + cdlen + 1 >= MAXNAMELEN) {
      return -E_BAD_PATH;
    }
    strcpy(dst, _curdir);
    if (dst[cdlen-1] != '/') {
      strcat(dst, "/");
    }
    strcat(dst, src);
  }
  return 0;
}

int
chdir(const char *dir) 
{
  char tmpdir[MAXNAMELEN+1];
  int r;

  if ((r = fix_dir(dir, tmpdir)) < 0) {
    return r;
  }

  if (!isdir(tmpdir))
    return -E_BAD_PATH;
      
  strcpy(_curdir, tmpdir);
  setenv("CWD", _curdir);
  return 0;
}

char *
getcwd(char *wd)
{
  fix_curdir();

  if (wd)
    strcpy(wd, _curdir);
  return _curdir;
}

int 
mkdir(const char *path, int mode) 
{
  struct Fd *fd;
  int r;
  char real_path[MAXPATHLEN];

  if ((r = fd_alloc(&fd)) < 0) {
    return r;
  }
  if ((r = sys_mem_alloc(0, (u_int)fd, PTE_P|PTE_U|PTE_W)) < 0) {
    return r;
  }

  fix_dir(path, real_path);
  if ((r = fsipc_open(real_path, mode|O_EXCL|O_CREAT, FTYPE_DIR, fd)) < 0){
    if (sys_mem_unmap(0, (u_int)fd) < 0)
      printf("open fsipc_open1 failed to unmap 0x%x\n", fd);
    return r;
  }
  file_close(fd);
  return 0;
}

int open(const char *path, int mode)
{ 
  return _internal_open(path, mode, FTYPE_REG);
}


int opendir(const char *path, int mode)
{
  return _internal_open(path, mode, FTYPE_DIR);
}

int isdir(const char *file)
{
  struct Stat st;
  int r;
  
  if ((r = stat(file, &st)) < 0) {
    return 0;
  }

  if (!st.st_isdir) {
    return 0;
  }

  return 1;
}

int isfile(const char *file)
{
  struct Stat st;
  int r;
  
  if ((r = stat(file, &st)) < 0) {
    return 0;
  }

  if (st.st_isdir) {
    return 0;
  }

  return 1;
}
