/*
 * Mark B. Hanson's malloc implementation, modified by me to work under 6.828 kernel
 */
#define MALLOC_H
#include "lib.h"

#undef OBLITERATE
#undef DISPLAY
#undef BOUNDS
#undef ZERO_ERROR

#define FALSE 0
#define TRUE 1

#define FIT(x) firstfit(x) /**/

#define PATTERN(p) ((char)((unsigned int)(p) & 0xff))
#define MAX(x,y) ((x) > (y) ? (x) : (y))

#define LOCATION(f,l) \
	if ((f) != NULL) { \
	    fprintf(stderr, " at `%s:%d'\n", (f), (l)); \
	} else { \
	    fprintf(stderr, "\n"); \
	}

#define ALIGN(p) (((p) + (sizeof(align_t) - 1)) & ~(sizeof(align_t) - 1))

#ifdef BOUNDS
#define ADDR(n) ((caddr_t)(n) + sizeof(node) + sizeof(align_t))
#define AFTER(n) (ADDR(n) + (n)->x.rsize + sizeof(align_t))
#define OVERHEAD (sizeof(node) + (2 * sizeof(align_t)))
#else
#define ADDR(n) ((caddr_t)(n) + sizeof(node))
#define AFTER(n) (ADDR(n) + (n)->x.rsize)
#define OVERHEAD (sizeof(node))
#endif /* BOUNDS */

#ifdef DISPLAY
#define STATUS(s) \
    fprintf(stderr, \
	"%d %d %d (%s)\n", \
	total, \
	alloc_len, \
	avail_len, \
	(s), \
        )
#else
#define STATUS(s)
#endif /* DISPLAY */

#define MINNODE (64 * sizeof(align_t))

#define ADJACENT(n1,n2) \
	((caddr_t)(n1) + OVERHEAD + (n1)->x.rsize == (caddr_t)(n2))


extern caddr_t sbrk();
extern int fputs();
extern int getpagesize();


typedef double align_t;

union _node {
    struct {
	size_t rsize; /* real size */
	size_t usize; /* user size */
	char *file;
	unsigned int line;
	union _node *next;
    } x;
    align_t align;
};
typedef union _node node;

enum _list {
    AVAILABLE,
    ALLOCATED
};
typedef enum _list list;


static node *available = NULL;
static node *allocated = NULL;
static unsigned long total = 0;
static unsigned long alloc_len = 0;
static unsigned long avail_len = 0;

int pagesize = 0;


/* put a breakpoint here for debugging */
static void debug() {
    total = total;
} /* debug */


#if 0
static void pnode(n)
    node *n;
{
    fprintf(stderr,
"0x%x: rsize = %u, usize = %u, addr = 0x%x, file = %s:%d, next = 0x%x\n",
	    (unsigned int)n,
	    n->x.rsize,
	    n->x.usize,
	    (unsigned int)ADDR(n),
	    n->x.file ? n->x.file : "unknown",
	    n->x.line,
	    (unsigned int)n->x.next);
} /* pnode */
#endif

#if 0
static void plist(head)
    node *head;
{
    node *c;

    fprintf(stderr, "\n=== list starting at 0x%x: ===\n", (unsigned int)head);

    c = head;
    while (c != NULL) {
	pnode(c);
	c = c->x.next;
    }
} /* plist */

#endif

static unsigned long llength(head)
    node *head;
{
    unsigned long i = 0;
    node *c;

    c = head;
    while (c != NULL) {
	i++;
	c = c->x.next;
    }
    return i;
} /* llength */


static void delete(l, n)
    list l;
    node *n;
{
    node *c, *p, **head;

    assert(l == ALLOCATED || l == AVAILABLE);
    assert(n != NULL);

    head = (l == ALLOCATED) ? &allocated : &available;
    assert(*head != NULL);

    assert(llength(allocated) == alloc_len);
    assert(llength(available) == avail_len);

    c = p = *head;
    while ((c != NULL) && (ADDR(c) > ADDR(n))) {
	p = c;
	c = c->x.next;
    }

    assert(c != NULL);

    if (l == ALLOCATED) {
	alloc_len--;
    } else {
	avail_len--;
    }

    if (c == p) {
	*head = c->x.next;
    } else {
	p->x.next = c->x.next;
    }
} /* delete */


static void insert(l, n)
    list l;
    node *n;
{
    node *c, *p, **head;

    assert(l == ALLOCATED || l == AVAILABLE);
    assert(n != NULL);

    head = (l == ALLOCATED) ? &allocated : &available;

    assert(llength(allocated) == alloc_len);
    assert(llength(available) == avail_len);

    c = p = *head;
    while ((c != NULL) && (ADDR(c) > ADDR(n))) {
	p = c;
	c = c->x.next;
    }

    assert(c == NULL || ADDR(c) != ADDR(n));

    if ((l == AVAILABLE) && (*head != NULL) && (c != NULL) && ADJACENT(c,n)) {
	c->x.rsize += OVERHEAD + n->x.rsize;
	if (ADJACENT(c,p)) {
	    delete(AVAILABLE, p);
	    c->x.rsize += OVERHEAD + p->x.rsize;
	}
    } else {
	if (l == ALLOCATED) {
	    alloc_len++;
	} else {
	    avail_len++;
	}
	n->x.next = c;
	if (c == p) {
	    *head = n;
	} else {
	    p->x.next = n;
	    if ((l == AVAILABLE) && (*head != NULL) && ADJACENT(n,p)) {
		delete(AVAILABLE, p);
		n->x.rsize += OVERHEAD + p->x.rsize;
	    }
	}
    }
} /* insert */


#ifdef BOUNDS
static void fillbounds(n)
    node *n;
{
    char *start;
    char pattern = PATTERN(n);
    int length, i;

    start = (char *)n + sizeof(node);
    length = sizeof(align_t);
    for (i = 0; i < length; i++) {
	start[i] = pattern;
    }

    start = (char *)n + sizeof(node) + sizeof(align_t) + n->x.usize;
    length = n->x.rsize - n->x.usize + sizeof(align_t);
    for (i = 0; i < length; i++) {
	start[i] = pattern;
    }
} /* fillbounds */


static void checkbounds(n, func)
    node *n;
    char *func;
{
    int i, start, end;
    unsigned char pattern = PATTERN(n);
    unsigned char current;
    
#define CHECK(i) \
	current = ADDR(n)[(i)]; \
	if (current != pattern) { \
	    fprintf(stderr, \
"\n%s: found `\\%03o' instead of `\\%03o' at 0x%x[%d].\nnode =", \
		    func, \
		    current, \
		    pattern, \
		    (unsigned int)ADDR(n), \
		    (i)); \
	    pnode(n); \
	    debug(); \
	}

    start = -sizeof(align_t);
    end = 0;
    for (i = start; i < end; i++) {
	CHECK(i);
    }

    start = n->x.usize;
    end = n->x.rsize + sizeof(align_t);
    for (i = start; i < end; i++) {
	CHECK(i);
    }
} /* checkbounds */


static void lbounds(head)
    node *head;
{
    node *c;

    c = head;
    while (c != NULL) {
	checkbounds(c, "lbounds");
	c = c->x.next;
    }
} /* lbounds */
#endif /* BOUNDS */


static node *find(head, ptr)
    node *head;
    caddr_t ptr;
{
    node *c;

    c = head;
    while((c != NULL) && (ADDR(c) > ptr)) {
	c = c->x.next;
    }

    if ((c == NULL) || (ADDR(c) != ptr)) {
	return NULL;
    }

    return c;
} /* find */


static node *expand(size)
    size_t size;
{
    node *n;
    size_t big;

#ifdef BOUNDS
    lbounds(allocated);
#endif /* BOUNDS */

    if (pagesize == 0) {
	pagesize = BY2PG;
    }

    big = MAX(8 * pagesize, (((size + OVERHEAD) / pagesize) + 1) * pagesize);

    assert(big >= size + OVERHEAD);

    n = (node *)sbrk(big);

    if (n == (node *)-1) {
	fprintf(stderr, "\nexpand: sbrk(%d) failed!\n", big);
	debug();
	return NULL;
    }
    n->x.rsize = big - OVERHEAD;
    n->x.next = NULL;
    n->x.file = __FILE__;
    n->x.line = __LINE__;

    return n;
} /* expand */


static node *firstfit(size)
    size_t size;
{
    node *c;
    size_t aligned = ALIGN(size);

#ifdef ZERO_ERROR
    assert(size != 0);
#endif /* ZERO_ERROR */

    c = available;
    while ((c != NULL) && (c->x.rsize < aligned)) {
	c = c->x.next;
    }

    if (c == NULL) {
	if ((c = expand(aligned)) == NULL) {
	    return NULL;
	}
    } else {
        assert(find(allocated, ADDR(c)) == NULL);
        assert(find(available, ADDR(c)) == c);

	delete(AVAILABLE, c);
    }

    assert(find(allocated, ADDR(c)) == NULL);
    assert(find(available, ADDR(c)) == NULL);

    if (c->x.rsize >= aligned + OVERHEAD + MINNODE) {
	node *n;
	size_t leftover;
	
	leftover = c->x.rsize - aligned - OVERHEAD;
	c->x.rsize = aligned;

	n = (node *)AFTER(c);
	n->x.rsize = leftover;

	assert(find(allocated, ADDR(n)) == NULL);
	assert(find(available, ADDR(n)) == NULL);

	insert(AVAILABLE, n);

	assert(find(allocated, ADDR(n)) == NULL);
	assert(find(available, ADDR(n)) == n);
    }

    c->x.usize = size;

#ifdef BOUNDS
    fillbounds(c);
#endif /* BOUNDS */

    insert(ALLOCATED, c);

    assert(find(allocated, ADDR(c)) == c);
    assert(find(available, ADDR(c)) == NULL);

    return c;
} /* firstfit */


#if 0
static node *bestfit(size)
    size_t size;
{
    node *c, *b;
    size_t aligned = ALIGN(size);
    size_t over, best = MAXINT;

#ifdef ZERO_ERROR
    assert(size != 0);
#endif /* ZERO_ERROR */

    b = NULL;
    c = available;
    while (c != NULL) {
	if (c->x.rsize >= size) {
	    over = c->x.rsize - size;
	    if ((over < best) || (c == available)) {
		best = over;
		b = c;
	    }
	}
	c = c->x.next;
    }

    if (b == NULL) {
	if ((b = expand(aligned)) == NULL) {
	    return NULL;
	}
    } else {
        assert(find(allocated, ADDR(b)) == NULL);
        assert(find(available, ADDR(b)) == b);

	delete(AVAILABLE, b);
    }

    assert(find(allocated, ADDR(b)) == NULL);
    assert(find(available, ADDR(b)) == NULL);

    if (b->x.rsize >= aligned + OVERHEAD + MINNODE) {
	node *n;
	size_t leftover;
	
	leftover = b->x.rsize - aligned - OVERHEAD;
	b->x.rsize = aligned;

	n = (node *)AFTER(b);
	n->x.rsize = leftover;

	assert(find(allocated, ADDR(n)) == NULL);
	assert(find(available, ADDR(n)) == NULL);

	insert(AVAILABLE, n);

	assert(find(allocated, ADDR(n)) == NULL);
	assert(find(available, ADDR(n)) == n);
    }

    b->x.usize = size;

#ifdef BOUNDS
    fillbounds(b);
#endif /* BOUNDS */

    insert(ALLOCATED, b);

    assert(find(allocated, ADDR(b)) == b);
    assert(find(available, ADDR(b)) == NULL);

    return b;
} /* bestfit */
#endif

/*---------------------------------------------------------------------------*/

caddr_t _malloc(size, file, line)
    size_t size;
    char *file;
    unsigned int line;
{
    node *store;

    if (size == 0) {
	fprintf(stderr, "\nmalloc: attempt to allocate 0 bytes");
	LOCATION(file, line);
#ifdef ZERO_ERROR
	if (file != NULL)
	    debug();
	return NULL;
#endif /* ZERO_ERROR */
    }

    store = FIT(size);

    if (store == NULL) {
	fprintf(stderr, "\nmalloc: unable to allocate %d bytes", size);
	LOCATION(file, line);
	debug();
	return NULL;
    }

#ifdef OBLITERATE
    memset(ADDR(store), '\001', size);
#endif /* OBLITERATE */

    store->x.file = file;
    store->x.line = line;

    total += size;
    STATUS("malloc ");

    return ADDR(store);
} /* _malloc */


caddr_t _calloc(num, size, file, line)
    unsigned int num;
    size_t size;
    char *file;
    unsigned int line;
{
    node *store;
    unsigned int bytes;

    if ((num == 0) && (size == 0)) {
	fprintf(stderr, "\ncalloc: attempt to allocate 0 items of size 0");
	LOCATION(file, line);
#ifdef ZERO_ERROR
	if (file != NULL)
	    debug();
	return NULL;
#endif /* ZERO_ERROR */
    }

    if (num == 0) {
	fprintf(stderr, "\ncalloc: attempt to allocate 0 items of size %d",
		size);
	LOCATION(file, line);
#ifdef ZERO_ERROR
	if (file != NULL)
	    debug();
	return NULL;
#endif /* ZERO_ERROR */
    }

    if (size == 0) {
	fprintf(stderr, "\ncalloc: attempt to allocate %d items of size 0",
		num);
	LOCATION(file, line);
#ifdef ZERO_ERROR
	if (file != NULL)
	    debug();
	return NULL;
#endif /* ZERO_ERROR */
    }

    bytes = num * size;

    store = FIT(bytes);

    if (store == NULL) {
	fprintf(stderr, "\ncalloc: unable to allocate %d bytes", bytes);
	LOCATION(file, line);
	debug();
	return NULL;
    }

    memset(ADDR(store), '\000', bytes);
    store->x.file = file;
    store->x.line = line;

    total += bytes;
    STATUS("calloc ");

    return ADDR(store);
} /* _calloc */


caddr_t _realloc(ptr, size, file, line)
    caddr_t ptr;
    size_t size;
    char *file;
    unsigned int line;
{
    node *t;
    node *store;

    if (ptr == NULL) {
	return _malloc(size, file, line);
    }

    t = find(allocated, ptr);

    if (t == NULL) {
	fprintf(stderr,
		"\nrealloc: attempt to realloc unallocated memory (0x%x)",
		(unsigned int)ptr);
	LOCATION(file, line);
	debug();
	return NULL;
    }

#ifdef BOUNDS
    checkbounds(t, "realloc");
#endif /* BOUNDS */

    total += size - t->x.usize;
    STATUS("realloc");

    if (t->x.rsize >= size) {

#ifdef OBLITERATE
	if (t->x.rsize != size) {
	    memset(ADDR(t) + t->x.usize, '\001', size - t->x.usize);
	}
#endif /* OBLITERATE */

	t->x.usize = size;
	t->x.file = file;
	t->x.line = line;

#ifdef BOUNDS
	fillbounds(t);
#endif /* BOUNDS */

	return ADDR(t);
    }

    store = FIT(size);

    if (store == NULL) {
	fprintf(stderr, "\nrealloc: unable to allocate %d bytes", size);
	LOCATION(file, line);
	debug();
	return NULL;
    }

    memcpy(ADDR(store), ADDR(t), MIN(t->x.usize, size));

    assert(find(available, ptr) == NULL);
    assert(find(allocated, ptr) == t);

    delete(ALLOCATED, t);
    insert(AVAILABLE, t);

    assert(find(allocated, ptr) == NULL);
    /*
    ** collapsing prevents us from checking:
    ** assert(find(available, ptr) == t);
    */

    store->x.file = file;
    store->x.line = line;

    return ADDR(store);
} /* _realloc */


void _free(ptr, file, line)
    caddr_t ptr;
    char *file;
    unsigned int line;
{
    node *t;

    if (ptr == NULL) {
	fprintf(stderr, "\nfree: attempt to free NULL");
	LOCATION(file, line);
	if (file != NULL)
	    debug();
	return;
    }

    t = find(allocated, ptr);

    if (t == NULL) {
	fprintf(stderr, "\nfree: attempt to free unallocated memory (0x%x)",
		(unsigned int)ptr);
	LOCATION(file, line);
	debug();
	return;
    }

#ifdef BOUNDS
    checkbounds(t, "free");
#endif /* BOUNDS */

    assert(find(allocated, ptr) == t);
    assert(find(available, ptr) == NULL);

    delete(ALLOCATED, t);
    insert(AVAILABLE, t);

    assert(find(allocated, ptr) == NULL);
    /*
    ** collapsing prevents us from checking:
    ** assert(find(available, ptr) == t);
    */

#ifdef OBLITERATE
    memset(ADDR(t), '\002', t->x.usize);
#endif /* OBLITERATE */

    total -= t->x.usize;
    STATUS("free   ");
} /* _free */

/*---------------------------------------------------------------------------*/
/*----------------------------- USER INTERFACE: -----------------------------*/
/*---------------------------------------------------------------------------*/

caddr_t malloc(size)
    size_t size;
{
    return _malloc(size, __FILE__, __LINE__);
} /* malloc */


caddr_t calloc(num, size)
    unsigned int num;
    size_t size;
{
    return _calloc(num, size, __FILE__, __LINE__);
} /* calloc */


caddr_t realloc(ptr, size)
    caddr_t ptr;
    size_t size;
{
    return _realloc(ptr, size, __FILE__, __LINE__);
} /* realloc */


void free(ptr)
    caddr_t ptr;
{
    _free(ptr, __FILE__, __LINE__);
} /* free */


void cfree(ptr)
    caddr_t ptr;
{
    _free(ptr, __FILE__, __LINE__);
} /* cfree */
